@@ -14,9 +14,9 @@ use clvm_tools_rs::classic::clvm_tools::stages::stage_0::DefaultProgramRunner;
 fn main() {
     let mut allocator = Allocator::new();
     let runner = Rc::new(DefaultProgramRunner::new());
-    let opts = Rc::new(DefaultCompilerOpts::new(&"*program*".to_string()));
+    let opts = Rc::new(DefaultCompilerOpts::new("*program*"));
     let stdin = io::stdin();
-    let mut repl = Repl::new(opts.clone(), runner.clone());
+    let mut repl = Repl::new(opts, runner);
 
     print!(">>> ");
     io::stdout().flush().unwrap();
@@ -29,7 +29,7 @@ fn main() {
                     .process_line(&mut allocator, line)
                     .map(|result| {
                         if let Some(result) = result {
-                            print!("{}\n>>> ", result.to_sexp().to_string());
+                            print!("{}\n>>> ", result.to_sexp());
                         } else {
                             print!("... ");
                         }
