@@ -1,10 +1,10 @@
-pub mod NodePath;
 pub mod binutils;
 pub mod clvmc;
 pub mod cmds;
 pub mod curry;
 pub mod debug;
 pub mod ir;
+pub mod node_path;
 pub mod pattern_match;
 pub mod sha256tree;
 pub mod stages;
